#include <stdio.h>
#include "lightbuffer.h"
#include "math.h"
#include "timing.h"

struct FColor {
  operator Color() {
    return Color(color[0], color[1], color[2]);
  }
  float color[3];
};

struct State {
  State() {}

  State(int n) : colors(n) {
    for (int i = 0; i < n; i++) {
      colors[i].color[0] = 0;
      colors[i].color[1] = 0;
      colors[i].color[2] = 0;
    }
  }
  std::vector<FColor> colors;

  void Add(int p, int c, float v) {
    v += colors[p].color[c];
    if (v < 0) v = 0;
    if (v > 255) v = 255;
    colors[p].color[c] = v;
  }

  void AddRGB(int p, float r, float g, float b) {
    if (p < 0)
      p = colors.size() - ((-p) % colors.size());
    else
      p %= colors.size();
    Add(p, 0, r);
    Add(p, 1, g);
    Add(p, 2, b);
  }

  void Decay(double multiplier) {
    // Decay lights
    for (int p = 0; p < colors.size(); p++) {
      AddRGB(p, -4 * multiplier, -4 * multiplier, -4 * multiplier);
    }
  }
};

State perm, temp;

class Sprite {
public:
  Sprite() : r(0), g(0), b(0), pos(0.0), speed(0.0) {
  }
  void Draw(double multiplier) {
    int p0 = (int)floor(pos);
    pos += speed * multiplier;
    int p = (int)floor(pos);
    double f = pos - p;
    temp.AddRGB(p + 0, 255 * (1 - f), 255 * (1 - f), 255 * (1 - f));
    temp.AddRGB(p + 1, 255, 255, 255);
    temp.AddRGB(p + 2, 255 * f, 255 * f, 255 * f);
    if (p0 < p) {
      for (int x = p0; x < p; x++) {
	perm.AddRGB(x, r, g, b);
      }
    } else {
      for (int x = p0; x > p; x--) {
	perm.AddRGB(x, r, g, b);
      }
    }
  }

  double pos, speed;
  int r, g, b;
};

int main(int argc, char **argv) {
  LightBuffer* l = GetLightBuffer();
  std::vector<int> geometry;
  l->GetGeometry(&geometry);
  int total = 0;
  for (int i = 0; i < geometry.size(); i++)
    total += geometry[i];

  perm = State(total);

  std::vector<Sprite> sprites(7);

  sprites[0].r = 255;
  sprites[1].g = 255;
  sprites[2].b = 255;
  sprites[3].r = sprites[3].g = 255;
  sprites[4].g = sprites[4].b = 255;
  sprites[5].b = sprites[5].r = 255;
  sprites[6].r = sprites[6].g = sprites[6].b = -255;
  sprites[0].speed = 0.2;
  sprites[1].speed = 0.5;
  sprites[2].speed = 0.7;
  sprites[3].speed = 1.1;
  sprites[4].speed = -1.0;
  sprites[5].speed = -0.6;
  sprites[6].speed = -0.3;

  double speed = 1.0/30;
  double z = 0.0;
  struct timeval start = Time();
  int frames = 0;
  l->Show();
  l->Show();
  l->Show();
  while (1) {
    perm.Decay(speed * 50);
    temp = perm;
    for (int i = 0; i < sprites.size(); i++) sprites[i].Draw(speed * 100);

    int n = 0;
    for (int i = 0; i < geometry.size(); i++) {
      for (int lamp = 0; lamp < geometry[i]; lamp++) {
	l->Set(i, lamp, temp.colors[n++]);
      }
    }
    l->Show();
    frames++;
    speed = (Time() - start) / frames;
  }
}
