#include <sys/select.h>
#include <sys/time.h>
#include <stdio.h>
#include "lights.h"
#include "math.h"

#define STRINGS 8
#define LIGHTS 50

typedef float Display[STRINGS][LIGHTS][3];

class IncrementalPainter : public TeensyLightWallRelay {
public:
  IncrementalPainter(const string &device, int num_strings) :
    TeensyLightWallRelay(device, num_strings) {
    for (int s = 0; s < STRINGS; s++) {
      for (int l = 0; l < LIGHTS; l++) {
	for (int c = 0; c < 3; c++) {
	  target_state_[s][l][c] = 0;
	  current_state_[s][l][c] = 0;
	  accum_err_[s][l][c] = 0;
	}
      }
    }
  }
  void Update() {
    unsigned char buffer[27];
    for (int i = 0; i < sizeof(buffer); i++) buffer[i] = 0;
    buffer[0] = 0xEE;
    for (int s = 0; s < STRINGS; s++) {
//      fprintf(stderr, " STRING = %d \n", s);
      double biggest_diff = -1.0;
      int update_l = 0;
      for (int l = 0; l < LIGHTS; l++) {
	double pixel_diff = 0;
	for (int c = 0; c < 3; c++) {
	  double diff = target_state_[s][l][c] - current_state_[s][l][c];
	  accum_err_[s][l][c] *= 0.5;
	  accum_err_[s][l][c] += diff;
	  pixel_diff += fabs(diff);
	}
	if (pixel_diff > biggest_diff) {
	  update_l = l;
	  biggest_diff = pixel_diff;
	}
      }
      int pos = 1;
      for (int i = 32; i; i >>= 1) {
	if (i & update_l)  buffer[pos] |= 1 << s;
	pos++;
      }
      for (int i = 128; i; i >>= 1) {
	if (i & 0xcc) buffer[pos] |= 1 << s;
	pos++;
      }
      for (int c = 0; c < 3 ; c++) {
	int color = (int)floor(target_state_[s][update_l][c] * 16);
#if 0
	fprintf(stderr, "c=%f e=%f t=%f\n",
		current_state_[s][update_l][c],
		accum_err_[s][update_l][c],
		target_state_[s][update_l][c]);
	fprintf(stderr, "Color = %d update_l = %d\n", color, update_l);
#endif
	if (color < 0) color = 0;
	if (color > 15) color = 15;
	current_state_[s][update_l][c] = color / 15.0;
	accum_err_[s][update_l][c] = 0.0;
//	fprintf(stderr, "nc=%f \n", current_state_[s][update_l][c]);
	for (int i = 8; i; i >>= 1) {
	  if (i & color) buffer[pos] |= 1 << s;
	  pos++;
	}
      }
    }
    if (write(fd_, buffer, sizeof(buffer)) != sizeof(buffer)) {
      perror("Write to Teensy.");
      exit(1);
    }
  }

public:
  Display target_state_;
private:
  Display current_state_;
  Display accum_err_;
};

int main() {
  IncrementalPainter relay("/dev/ttyACM0", 8);
  vector<LampColor> colors(8);

  relay.Init();

  double z = 0.0;
  int frame = 0;
  while (1) {
    for (int lamp = 0; lamp < 50; lamp++) {
      for (int i = 0; i < 8; i++) {
	relay.target_state_[i][lamp][0] = sin(z + lamp / 4.0)/2 + 0.5;
	relay.target_state_[i][lamp][1] = sin(z * 0.77 + lamp / 4.0)/2 + 0.5;
	relay.target_state_[i][lamp][2] = sin(z * 0.55 + lamp / 4.0)/2 + 0.5;
      }
    }
    relay.Update();
    z += 0.2 / 50.0;
    fprintf(stderr, "\rFrame = %d ", frame++);
  }
}
