#include <stdlib.h>
#include <iostream>
#include "matrix.h"
#include "timing.h"

#define ASSERT_EQ(T,A,B) do {			\
  T a_ = (A);					\
  T b_ = (B);					\
  if( !( a_ == b_) ) {				\
    std::cout << "ASSERT_EQ failed.\n";		\
    std::cout << "a = " << a_ << "\n";		\
    std::cout << "b = " << b_ << "\n";		\
    exit(0);					\
  }						\
}while(0)

void multiplication_test() {
  Matrix<double> a(1,2);
  a.set(0, 0, 3);
  a.set(0, 1, 5);

  Matrix<double> b(2,1);
  b.set(0, 0, 7);
  b.set(1, 0, 11);

  Matrix<double> c(1,1);
  c.set(0, 0, 76);

  ASSERT_EQ(Matrix<double>, b*a, c);

  Matrix<double> d(2,2);
  d.set(0, 0, 21);
  d.set(1, 0, 33);
  d.set(0, 1, 35);
  d.set(1, 1, 55);

  ASSERT_EQ(Matrix<double>, a*b, d);
}

void stupid_solve_test() {
  Matrix<double> a(1, 1);
  a.set(0, 0, 3);
  Matrix<double> b(1, 1);
  b.set(0, 0, 9);
  Matrix<double> c(1, 1);
  c.set(0, 0, 3);
  ASSERT_EQ(bool, a.solve(b), true);
  ASSERT_EQ(Matrix<double>, b, c);
}

void solve_test() {
  Matrix<double> a(2,2);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);

  Matrix<double> b(1,2);
  b.set(0, 0, 7);
  b.set(0, 1, 11);

  Matrix<double> c(1,2);
  c.set(0, 0, -3);
  c.set(0, 1, 5);

  ASSERT_EQ(bool, a.solve(b), true);
  ASSERT_EQ(Matrix<double>, b, c);
}

void solve_test2() {
  Matrix<double> a(2,4);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);
  a.set(0, 2, 5);
  a.set(1, 2, 6);
  a.set(0, 3, 7);
  a.set(1, 3, 8);

  Matrix<double> b(1,4);
  b.set(0, 0, 7);
  b.set(0, 1, 11);
  b.set(0, 2, 17);
  b.set(0, 3, 13);

  Matrix<double> c(1,2);
  c.set(0, 0, -6);
  c.set(0, 1, 7.2);

  Matrix<double> d(2,2);
  d.set(0, 0, 84);
  d.set(1, 0, 100);
  d.set(0, 1, 100);
  d.set(1, 1, 120);
  ASSERT_EQ(Matrix<double>, a.transpose() * a, d);
  ASSERT_EQ(Matrix<double>, a.solve_least_square(b), c);
}

void nnls_test() {
  Matrix<float> a(2,4);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);
  a.set(0, 2, 5);
  a.set(1, 2, 6);
  a.set(0, 3, 7);
  a.set(1, 3, 8);

  Matrix<float> b(1,4);
  b.set(0, 0, 7);
  b.set(0, 1, 11);
  b.set(0, 2, 17);
  b.set(0, 3, 13);

  Matrix<float> c(1,2);
  c.set(0, 0, 0.0);
  c.set(0, 1, 2.2);

  ASSERT_EQ(Matrix<float>, solve_nnls(a, b), c);
}

void fnnls_test() {
  Matrix<float> a(2,4);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);
  a.set(0, 2, 5);
  a.set(1, 2, 6);
  a.set(0, 3, 7);
  a.set(1, 3, 8);

  Matrix<float> b(1,4);
  b.set(0, 0, 7);
  b.set(0, 1, 11);
  b.set(0, 2, 17);
  b.set(0, 3, 13);

  Matrix<float> c(1,2);
  c.set(0, 0, 0.0);
  c.set(0, 1, 2.2);

  ASSERT_EQ(Matrix<float>, a.fnnls(b), c);
}

void fnnls_test2() {
  Matrix<float> a(4,3);
  a.set(0, 0, 10);
  a.set(0, 1, 1);
  a.set(0, 2, 10);
  a.set(1, 0, 20);
  a.set(1, 1, 20);
  a.set(1, 2, 100);
  a.set(2, 0, 30);
  a.set(2, 1, 3);
  a.set(2, 2, 10);
  a.set(3, 0, 0);
  a.set(3, 1, 6);
  a.set(3, 2, 10);

  Matrix<float> b(1,3);
  b.set(0, 0, 6);
  b.set(0, 1, 3);
  b.set(0, 2, 2);

  ASSERT_EQ(Matrix<float>, a.fnnls(b), solve_nnls(a, b));
}

void cholesky_test() {
  Matrix<float> a(3, 3);
  a.set(0, 0, 25);
  a.set(0, 1, 15);
  a.set(1, 0, 15);
  a.set(0, 2, -5);
  a.set(2, 0, -5);
  a.set(1, 1, 18);
  a.set(2, 2, 11);

  Matrix<float> b(3, 3);
  b.set(0, 0, 5);
  b.set(1, 1, 3);
  b.set(2, 2, 3);
  b.set(0, 1, 3);
  b.set(0, 2, -1);
  b.set(1, 2, 1);
  ASSERT_EQ(Matrix<float>, a.cholesky(), b);
}


#if 0
void owlqn_nnls_test() {
  Matrix<float> a(2,4);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);
  a.set(0, 2, 5);
  a.set(1, 2, 6);
  a.set(0, 3, 7);
  a.set(1, 3, 8);

  Matrix<float> b(1,4);
  b.set(0, 0, 7);
  b.set(0, 1, 11);
  b.set(0, 2, 17);
  b.set(0, 3, 13);

  Matrix<float> c(1,2);
  c.set(0, 0, 0.0);
  c.set(0, 1, 2.2);

  ASSERT_EQ(Matrix<float>, solve_nnls_owlqn(a, b), c);
}
#endif

#if 0
Matrix<float> blocknnls(Matrix<float> A, Matrix<float> b) {
  int m = A.rows();
  int n = A.columns();
  assert(A.rows() == b.rows());
  assert(b.columns() == 1);

  F = []; 
  G = 1:n;
  x = zeroes(n, 1);
  Matrix<float> Atb = A.transpose() * b;
  Matrix<float> y = - Atb;
  p = 10;
  ninf = n + 1;
  alpha = 1:n;
  bool noready = true;
  while (noready) {
    // Step 1
    xF = x(F);
    yG = y(G);
    if (all(xF>=0) & all(yG>=0)) {
      x = zeros(n,1);
      y = zeros(n,1);
      x(F) = xF(1:length(F));
      break;
    } else {
      H1 = F(xF < 0);
      H2 = G(yG < 0);
      H = union(H1,H2);
      if (length(H) < ninf) {
	ninf = length(H);
      } else {
	if (p >= 1) {
	  p = p - 1;
	} else {
	  // p==0
                
	  r = max(H);

	  if (ismember(r, H1)) {
	    H1 = r;
	    H2 = [];
	  } else {
	    H1 = [];
	    H2 = r;
	  }
	}
      }
      F = union(setdiff(F,H1),H2);
      G = union(setdiff(G,H2),H1);
    }
    // Step 2
    AF = A(:,F);
    AG = A(:,G);
    xF = AF\b;
    yG = AG.transpose()*(AF*xF-b);
    x(F) = xF;
    y(G) = yG;
  }
  return x;
}

#endif

#if 1
Matrix<float> colsel(const Matrix<float>& a, const std::vector<bool>& c) {
  std::vector<int> index;
  for (int i = 0; i < c.size(); i++) {
    if (c[i]) {
      index.push_back(i);
    }
  }
  Matrix<float> ret(index.size(), a.rows());
  for (int y = 0; y < a.rows(); y++) {
    for (int x = 0; x < index.size(); x++) {
      ret.set(x, y, a.get(index[x], y));
    }
  }
  return ret;
}

void Splice(Matrix<float>& b, const std::vector<bool>& p, const Matrix<float>& sp) {
  int j = 0;
  for (int i = 0 ; i < p.size(); i++)
    if (p[i])
      b.set(0, i, sp.get(0, j++));
}

float FindMax(const Matrix<float>& x, int& index) {
  assert(x.columns() == 1);
  float ret = x.get(0,0);
  index = 0;
  for (int y = 1; y < x.rows(); y++) {
    if (x.get(0, y) > ret) {
      ret = x.get(0, y);
      index = y;
    }
  }
  return ret;
}

bool anyfalse(const std::vector<bool>& p) {
  for (int i = 0; i < p.size(); i++) if (p[i] == false) return true;
  return false;
}

Matrix<float> fastnnls(Matrix<float> x, Matrix<float> y, Matrix<float> b) {
//  function b = fastnnls(x,y,tol,b) 
  int m = x.rows();
  int n = x.columns();
  assert(y.columns() == 1);
  assert(b.columns() == 1);
  assert(y.rows() == m);
  assert(b.rows() == n);
  double tol = pow(2.0, -52.0) * x.norm(100000) * std::max<int>(m, n);
  std::vector<bool> p(n);
  for (int i = 0; i < n; i++) p[i] = b.get(0, i) > 0;
  // r = ~p;

  Matrix<float> sp = colsel(x, p).solve_least_square(y);
  Splice(b, p, sp);

  while (sp.min() < 0) {
    for (int i = 0; i < n; i++) {
      if (b.get(0, i) < 0) b.set(0, i, 0);
      p[i] = b.get(0, i) > 0;
    }
    sp = colsel(x, p).solve_least_square(y);
    Splice(b, p, sp);
  }
 
  Matrix<float> w = x.transpose() * (y - x*b); 

  int ind;
  float wmax = FindMax(w, ind);
  while (wmax > tol && anyfalse(p)) {
    p[ind] = true;
    sp = colsel(x, p).solve_least_square(y);
    while (sp.min() < 0) {
      Matrix<float> tsp(1, n);
      Splice(tsp, p, sp);

      float alpha = 0.0;
      for (int i = 0; i < n; i++) {
	if (b.get(0, i) != 0.0) {
	  float rat = b.get(0, i) / ( b.get(0, i) - tsp.get(0, i) );
	  if (rat > 0) {
	    if (alpha == 0.0)
	      alpha = rat;
	    else
	      alpha = std::min(rat, alpha);
	  }
	}
      }
      b += (tsp - b) * alpha;
      for (int i = 0; i < m; i++) p[i] = b.get(0, i) > 0;
      sp = colsel(x, p).solve_least_square(y);
    }
    Splice(b, p, sp);
    w = x.transpose()*(y-x*b);
    wmax = FindMax(w, ind);
    if (p[ind]) wmax = 0; 
  }

  return b;
}

void fastnnls_test() {
  Matrix<float> a(2,4);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);
  a.set(0, 2, 5);
  a.set(1, 2, 6);
  a.set(0, 3, 7);
  a.set(1, 3, 8);

  Matrix<float> b(1,4);
  b.set(0, 0, 7);
  b.set(0, 1, 11);
  b.set(0, 2, 17);
  b.set(0, 3, 13);

  Matrix<float> x(1, 2);

  Matrix<float> c(1,2);
  c.set(0, 0, 0.0);
  c.set(0, 1, 2.2);

  ASSERT_EQ(Matrix<float>, fastnnls(a, b, x), c);
}

void fastnnls_test2() {
  Matrix<float> a(4,3);
  a.set(0, 0, 10);
  a.set(0, 1, 1);
  a.set(0, 2, 10);
  a.set(1, 0, 20);
  a.set(1, 1, 20);
  a.set(1, 2, 100);
  a.set(2, 0, 30);
  a.set(2, 1, 3);
  a.set(2, 2, 10);
  a.set(3, 0, 0);
  a.set(3, 1, 6);
  a.set(3, 2, 10);

  Matrix<float> b(1,3);
  b.set(0, 0, 6);
  b.set(0, 1, 3);
  b.set(0, 2, 2);

  Matrix<float> x(1, 4);

  ASSERT_EQ(Matrix<float>, a.fnnls(b), fastnnls(a, b, x));
}

#endif


void matrix_bench() {
  Matrix<float> a(100,10000);
  Matrix<float> b(1,10000);
  for (int x = 0; x < 10000; x++) {
    for (int y = 0; y < 100; y++) {
      a.set(y, x,  (x * 97 + y * 33) % 87);
    }
    b.set(0, x, sin(x) + 3.0);
  }
  struct timeval start = Time();
  int iterations = 0;
  while (Time() - start < 10.0) {
    solve_nnls(a, b);
    iterations++;
  }
  fprintf(stderr, "NNLS: %f solves / s\n", iterations / (Time() - start));
#if 1
  start = Time();
  iterations = 0;
  while (Time() - start < 10.0) {
    a.fnnls(b);
    iterations++;
  }
  fprintf(stderr, "FNNLS: %f solves / s\n", iterations / (Time() - start));
#endif
#if 1
  start = Time();
  iterations = 0;
  while (Time() - start < 10.0) {
    Matrix<float> c(1, 100);
    fastnnls(a, b, c);
    iterations++;
  }
  fprintf(stderr, "FASTNNLS: %f solves / s\n", iterations / (Time() - start));
#endif
}

#define RUN(test) do {				\
  fprintf(stderr, "Running: %s\n", #test);	\
  test();					\
  fprintf(stderr, "%s: PASS\n", #test);		\
} while(0)
  

int main(int argc, char **argv) {
  RUN(multiplication_test);
  RUN(stupid_solve_test);
  RUN(solve_test);
  RUN(solve_test2);
  RUN(nnls_test);
  RUN(fnnls_test);
  RUN(fnnls_test2);
  RUN(fastnnls_test);
  RUN(fastnnls_test2);
  RUN(cholesky_test);
//  RUN(owlqn_nnls_test);
  RUN(matrix_bench);
}
