#ifndef NOT_RANDOM_H
#define NOT_RANDOM_H

#include "lights.h"

#define LIGHT_BATCHES 4

class NotRandom {
public:
  typedef unsigned long long u64;

  static u64 powm(u64 a, u64 power, u64 mod) {
    int ret = 1;
    // ret * a ^ power, reduce power to zero
    while (power) {
      if (power & 1) {
	ret = (ret * a) % mod;
	power--;
      }
      a = (a * a) % mod;
      power >>= 1;
    }
    return ret;
  }

  static u64 Scramble(u64 x) {
    return powm(x, 7919LLU, 15485863LLU);
  }
  
  static bool RandomOn(int frame, int light_string, int light) {
    if (light == -1) return true;
    if ((light % LIGHT_BATCHES) != (frame % LIGHT_BATCHES)) return false;
    u64 unique = frame;
    unique <<= 16;
    unique |= light_string;
    unique <<= 16;
    unique |= light;
    return (Scramble(unique) & 1) == 1;
  }

  static unsigned char RandomLevel(int frame, int light_string, int light) {
    if (light == -1) return 0;
    u64 unique = frame;
    unique <<= 16;
    unique |= light_string;
    unique <<= 16;
    unique |= light;
    unsigned char x = Scramble(unique) & 0xff;
    return (x * x) >> 8;
  }


};

#endif  // NOT_RANDOM_H
