#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <algorithm>

#include "matrix.h"
#include "random.h"
#include "images.h"
#include "thread_helper.h"
#include "lightbuffer.h"

int lights;
int num_images;
std::vector<int> geometry;
Matrix<float> per_light[10000];
Matrix<float> base;

int compare(const vector<unsigned long long>& a,
	    const vector<unsigned long long>& b) {
  int ret = 0;
  if (a.size() != b.size()) return -1;
  for (int i = 0; i < a.size(); i++) {
    ret += __builtin_popcount(a[i] ^ b[i]);
  }
  return ret;
}

class SolvePixelsThreader : public Threader {
public:
  SolvePixelsThreader(Matrix<float> *images) : images_(images) {
    int xsize = images_[0].columns();
    int ysize = images_[0].rows(); 
    
    for (int l = 0; l < lights; l++) {
      per_light[l] = Matrix<float>(xsize, ysize);
    }
    base = Matrix<float>(xsize, ysize);

    Run(images_[0].columns());
  }
  
  void DoTask(int x) {
    int xsize = images_[0].columns();
    int ysize = images_[0].rows(); 
#if 0
    const size_t S = sizeof(unsigned long long)*8;
    vector<vector<unsigned long long> > bit_vectors(strings * lights);
    for (int s = 0; s < strings; s++) {
      for (int l = 0; l < lights; l++) {
	bit_vectors[s * lights + l].resize((num_images + S) / S);
	for (int image = 0; image < num_images; image++) {
	  if (NotRandom::RandomOn(image, s, l)) {
	    bit_vectors[s * lights + l][image / S] |= 1ULL << (image % S);
	  }
	}
      }
    }
    
    for (int y = 0; y < ysize; y++) {
      std::vector<float> tmp;
      for (int image = 0; image < num_images; image++) {
	tmp.push_back(images_[image].get(x, y));
      }
      int n = num_images / 2;
      nth_element(tmp.begin(), tmp.begin() + n, tmp.end());
      float threshold = tmp[n];
      vector<unsigned long long> bits((num_images + S) / S);
      for (int image = 0; image < num_images; image++) {
	if (images_[image].get(x, y) >= threshold) {
	  if (x == 130 && y == 56) {
	    fprintf(stderr, "Image %d, ON\n", image);
	  }
	  bits[image / S] |= 1ULL << (image % S);
	}
      }

      int best_diff = num_images / 2;
      for (int s = 0; s < strings; s++) {
	for (int l = 0; l < lights; l++) {
	  int diff = compare(bits, bit_vectors[s * lights + l]);
	  if (diff < best_diff) {
	    if (x == 130 && y == 56) {
	      fprintf(stderr, "s = %d  l = %d  diff = %d threshold = %f\n", s, l, diff, threshold);
	    }
	    per_light[s][l].set(x, y, 1.0);
	    best_diff = diff;
	  }
	}
      }
    }
#else
    std::vector<Matrix<float> > equations(LIGHT_BATCHES);
    for (int batch = 0; batch < LIGHT_BATCHES; batch++) {
      int row = 0;
      int col = 1;
      for (int image = batch; image < num_images; image += LIGHT_BATCHES) {
	row++;
      }
      for (int s = 0; s < geometry.size(); s++) {
	for (int l = batch; l < geometry[s]; l += LIGHT_BATCHES) {
	  col++;
	}
      }
      equations[batch] = Matrix<float>(col, row);
    }

    for (int batch = 0; batch < LIGHT_BATCHES; batch++) {
      int row = 0;
      for (int image = batch; image < num_images; image += LIGHT_BATCHES) {
	int col = 0;
	
	for (int s = 0; s < geometry.size(); s++) {
	  for (int l = batch; l < geometry[s]; l += LIGHT_BATCHES) {
	    assert(!NotRandom::RandomOn(image, s, l+1));
	    assert(!NotRandom::RandomOn(image+1, s, l));
	    equations[batch].set(col++, row, NotRandom::RandomOn(image, s, l));
	  }
	}
	equations[batch].set(col++, row, 1.0);
	row++;
      }
    }

    for (int batch = 0; batch < LIGHT_BATCHES; batch++) {
      for (int y = 0; y < ysize; y++) {
	Matrix<float> ans(1, equations[batch].rows());
	int row = 0;
	for (int image = batch; image < num_images; image += LIGHT_BATCHES) {
	  ans.set(0, row, images_[image].get(x, y));
	  row++;
	}
	Matrix<float> ret = solve_nnls(equations[batch], ans);
#if 0
	FILE* f = fopen("A.matix", "w");
	save_matrix(equations[batch], f);
	fclose(f);
	f = fopen("B.matix", "w");
	save_matrix(ans, f);
	fclose(f);
	f = fopen("C.matix", "w");
	save_matrix(ret, f);
	fclose(f);
	exit(1);
#endif	
	
	int col = 0;
	int string_start = 0;
	for (int s = 0; s < geometry.size(); s++) {
	  for (int l = batch; l < geometry[s]; l += LIGHT_BATCHES) {
	    float v = ret.get(0, col);
//	  if (v < 0.025) v = 0.0;
	    per_light[string_start + l].set(x, y, v);
	    col++;
	  }
	  string_start += geometry[s];
	}
	base.set(x, y, ret.get(0, col));
      }
    }

#if 0
    for (int y = 0; y < ysize; y++) {
      Matrix<float> ans(1, num_images);
      for (int image = 0; image < num_images; image++) {
	ans.set(0, image, images_[image].get(x, y));
      }
      Matrix<float> ret = solve_nnls(eq, ans);
//      Matrix<float> ret = eq.fnnls(ans);
//      Matrix<float> ret = solve_nnls_owlqn(eq, ans);
      
      int col = 0;
      for (int l = 0; l < lights; l++) {
#if 0
	per_light[l].value(x, y) =
	  max<double>(0.01, ret.value(0, col++));
#endif
#if 0
	per_light[l].set(x, y, ret.get(0, col++));
#endif
#if 1
	float v = ret.get(0, col++);
//	  if (v < 0.025) v = 0.0;
	per_light[l].set(x, y, v);
#endif

	}
      base.set(x, y, ret.get(0, col));
    }
#endif
#endif
  }

private:
  Matrix<float> *images_;
};

class ReadImageThreader : public Threader{
public:
  ReadImageThreader(Matrix<float> *images, int num_images) :
    images_(images),
    num_images_(num_images) {
    Run(num_images);
  }

  void DoTask(int x) {
    images_[x] = read_image(x);
  }
private:
  Matrix<float>* images_;
  int num_images_;
};

void separate(int total_lights, int num_images) {
  Matrix<float> images[num_images];
  // for (int i = 0; i < num_images; i++) images[i] = read_image(i);
  ReadImageThreader image_reader(images, num_images);
  int xsize = images[0].columns();
  int ysize = images[0].rows();
  fprintf(stderr, "xsize = %d ysize = %d\n", xsize, ysize);
  // solve_pixels(images);
  SolvePixelsThreader solve_pixels(images);
  fprintf(stderr, "Saving...");

  FILE *f=fopen("light-images/base.pgm", "w");
  write_pgm(f, base);
  fclose(f);

  f=fopen("light-matrices/base.matrix", "w");
  save_matrix(base, f);
  fclose(f);

  for (int l = 0; l < total_lights; l++) {
    char filename[512];
    sprintf(filename,"light-images/%06d.pgm", l);
    f=fopen(filename, "w");
    write_pgm(f, per_light[l]);
    fclose(f);
    
    sprintf(filename,"light-matrices/%06d.matrix", l);
    f=fopen(filename, "w");
    save_matrix(per_light[l], f);
    fclose(f);
    if (!(l & 0xf)) fprintf(stderr, ".");
  }
  fprintf(stderr, "Done\n");
}

int main(int argc, char **argv) {
  LightBuffer *lb = GetLightBufferStdout();
  lb->GetGeometry(&geometry);
  for (int i = 0; i < geometry.size(); i++) {
    lights += geometry[i];
  }
  num_images = atoi(argv[1]);
  separate(lights, num_images);
}
