#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "matrix.h"
#include "random.h"
#include "images.h"
#include "lightbuffer.h"


int main(int argc, char **argv) {
  LightBuffer* lb = GetLightBufferStdout();
  std::vector<int> geometry;
  lb->GetGeometry(&geometry);
  std::vector<int> string_start(geometry.size());
  int lights = 0;
  for (int i = 0; i < geometry.size(); i++) {
    string_start[i] = lights;
    lights += geometry[i];
  }

  Matrix<float> per_light[lights];
  for (int l = 0; l < lights; l++) {
    fprintf(stderr, "\r Reading image %d ...", l);
    char filename[512];
    sprintf(filename,"light-images/%06d.pgm", l);
    FILE *f=fopen(filename, "r");
    per_light[l] = readpgm(f, 1);
    fclose(f);
  }
  fprintf(stderr, "\n");
  int xsize = per_light[0].columns();
  int ysize = per_light[0].rows();

  Matrix<float> rgb[3];
  int frame = 0;

  while (1) {
    switch (getchar()) {
      case EOF: exit(0);
      default:
	fprintf(stderr, "Whut kind of file is this???\n");
	exit(1);
      case 0xEE:
        for (int c = 0; c < 3; c++) rgb[c] = Matrix<float>(xsize, ysize);

	for (int s = 0; ; s++) {
	  int lights = getchar();
	  if (lights == EOF) {
	    fprintf(stderr, "End of file!\n");
	    exit(1);
	  }
	  lights <<= 8;
	  lights |= (unsigned char)getchar();
	  if (lights == 0) break;
	  for (int l = 0; l < lights; l++) {
            int RGB[3];
	    RGB[0] = getchar();
	    if (RGB[0] == EOF) {
	      fprintf(stderr, "End of file!\n");
	      exit(1);
	    }
	    RGB[1] = getchar();
	    RGB[2] = getchar();
            if (l >= geometry[s]) continue;

            for (int c = 0; c < 3; c++) {
              if (!RGB[c]) continue;
              rgb[c] += per_light[string_start[s] + l] * (RGB[c] / 255.0);
            }
	  }
	}
	fprintf(stderr, "\rWriting frame %d ", frame++);
        write_pnm(stdout, rgb);
    }
  }
}
