#include <stdio.h>
#include <stdlib.h>
#include "lightbuffer.h"
#include "math.h"
#include "timing.h"

struct FColor {
  operator Color() {
    return Color(color[0], color[1], color[2]);
  }
  float color[3];
};

struct State {
  State() {}

  State(int n) : colors(n) {
    for (int i = 0; i < n; i++) {
      colors[i].color[0] = 0;
      colors[i].color[1] = 0;
      colors[i].color[2] = 0;
    }
  }
  std::vector<FColor> colors;

  void Add(int p, int c, float v) {
    v += colors[p].color[c];
    if (v < 0) v = 0;
    if (v > 255) v = 255;
    colors[p].color[c] = v;
  }

  void AddRGB(int p, float r, float g, float b) {
    if (p < 0)
      p = colors.size() - ((-p) % colors.size());
    else
      p %= colors.size();
    Add(p, 0, r);
    Add(p, 1, g);
    Add(p, 2, b);
  }

  void Decay(double multiplier) {
    // Decay lights
    for (int p = 0; p < colors.size(); p++) {
      AddRGB(p, -4 * multiplier, -4 * multiplier, -4 * multiplier);
    }
  }
};

State perm;


int main(int argc, char **argv) {
  LightBuffer* l = GetLightBuffer();
  std::vector<int> geometry;
  l->GetGeometry(&geometry);
  int total = 0;
  for (int i = 0; i < geometry.size(); i++)
    total += geometry[i];

  perm = State(total);

  double speed = 1.0/30;
  double z = 0.0;
  struct timeval start = Time();
  int frames = 0;
  l->Show();
  l->Show();
  l->Show();
  while (1) {
    perm.Decay(speed * 100);
    for (int x = 0; x < 1 + speed * 1000; x++) {
      int c = rand();
      // Random colors
      perm.AddRGB(rand(), c & 255, (c>>8) & 255, (c>>16) & 255);

      // White
      //      perm.AddRGB(rand(), 255,255,255);
    }

    int n = 0;
    for (int i = 0; i < geometry.size(); i++) {
      for (int lamp = 0; lamp < geometry[i]; lamp++) {
	l->Set(i, lamp, perm.colors[n++]);
      }
    }
    l->Show();
    frames++;
    speed = (Time() - start) / frames;
  }
}
