#include <stdio.h>
#include "lights.h"
#include "math.h"

int kernel[3][3][3] = {
  { { 1, 1, -1 }, { 1, 1, 1 }, { -2, 1, 1 }, },
  { { 1, 1, 1 }, { 1, 0, 1 }, { 3, 1, 1 }, },
  { { 1, 1, 1 }, { 1, 1, 5 }, { 1, 1, 1 }, }
};

class State {
public:
  State() {
    for (int s=0; s < 8; s++) {
      for (int l=0; l < 50; l++) {
	for (int c=0; c < 3; c++) {
	  lights[s][l][c] = rand() & 255;
	}
      }
    }
  }
  void UpdateFrom(State *other) {
    for (int s=0; s < 8; s++) {
      for (int l=0; l < 50; l++) {
	for (int c=0; c < 3; c++) {
	  lights[s][l][c] = other->getKernelSum(s,l,c);
	}
      }
    }
  }

  int getKernelSum(int s, int l, int c) {
    int sum = 0;
    for (int ds = -1; ds <= 1; ds++) {
      for (int dl = -1; dl <= 1; dl++) {
	for (int dc = -1; dc <= 1; dc++) {
	  sum += kernel[ds+1][dl+1][dc+1] * get(s+ds,l+dl,c+dc);
	}
      }
    }
    sum /= 27;
    sum &= 255;
    return sum;
  }

  int get(int s, int l, int c) {
    return lights[s % 8][l % 50][c % 3];
  }

  void SetState(TeensyLightWallRelay &relay) {
    vector<LampColor> colors(8);
    for (int lamp = 0; lamp < 50; lamp++) {
      for (int i = 0; i < 8; i++) {
	colors[i].intensity = 0xFF;
	colors[i].color[0] = lights[i][lamp][0];
	colors[i].color[1] = lights[i][lamp][1];
	colors[i].color[2] = lights[i][lamp][2];
      }
      relay.SetPixels(lamp, colors);
    }
  }

  int lights[8][50][3];
};

int main() {
  TeensyLightWallRelay relay("/dev/ttyACM0", 8);

  relay.Init();

  State s1;
  State s2;

  int frame = 0;
  while (1) {
    s1.UpdateFrom(&s2);
    s1.SetState(relay);

    fprintf(stderr, "\rFrame = %d ", frame++);

    s2.UpdateFrom(&s1);
    s1.SetState(relay);
    fprintf(stderr, "\rFrame = %d ", frame++);
  }
}
