#include <stdio.h>
#include "lightbuffer.h"
#include "math.h"
#include "timing.h"

struct FloatColor {
  FloatColor() : r(0), g(0), b(0) {}
  FloatColor(float r, float g, float b) : r(r), g(g), b(b) {}
  Color GetColor() {
    return Color(std::max(0, std::min(255, (int)r)),
		 std::max(0, std::min(255, (int)g)),
		 std::max(0, std::min(255, (int)b)));
  }
  FloatColor operator*(float x) const { return FloatColor(r*x, g*x, b*x); }
  void operator+=(const FloatColor& c) { r+=c.r; g+=c.g; b+=c.b; }
  FloatColor operator+(const FloatColor& c) const { return FloatColor(r+c.r, g+c.g, b+c.b); }
  void Reset() { r=g=b=0.0; }

  float r, g, b;
};

FloatColor Transform(FloatColor c) {
  return FloatColor(cos(c.r)*cos(c.g),
		    cos(c.b)*sin(c.g) + sin(c.b)*sin(c.r)*cos(c.g),
		    sin(c.b)*sin(c.g) - cos(c.b)*sin(c.r)*cos(c.g));
}

int main(int argc, char **argv) {
  LightBuffer* l = GetLightBuffer();
  std::vector<int> geometry;
  l->GetGeometry(&geometry);

  double z = 0.0;
  int frame = 0;
  struct timeval t = Time();
  for (int i = 0; i < geometry.size(); i++) {
    for (int lamp = 0; lamp < geometry[i]; lamp++) {
      l->Set(i, lamp, Color(0,0,0));
    }
  }
  l->Show();
  l->Show();
  l->Show();
  while (1) {
    int j = 0;
    for (int i = 0; i < geometry.size(); i++) {
      for (int lamp = 0; lamp < geometry[i]; lamp++) {
	++j;
	FloatColor a = Transform(FloatColor(z + j/137.0, z/2 + j/177.0, -z / 3 + j/2));
	l->Set(i, lamp, (FloatColor(128, 128, 128) + a * 200).GetColor());
      }
    }
    l->Show();

    struct timeval now = Time();

    z += 5.0 * (now - t);
    t = now;
    fprintf(stderr, "\rFrame = %d ", frame++);
  }
}
