#include <stdio.h>
#include "lightbuffer.h"
#include "math.h"
#include "timing.h"

struct FloatColor {
  FloatColor() : r(0), g(0), b(0) {}
  FloatColor(float r, float g, float b) : r(r), g(g), b(b) {}
  Color GetColor() {
    return Color(std::max(0, std::min(255, (int)r)),
		 std::max(0, std::min(255, (int)g)),
		 std::max(0, std::min(255, (int)b)));
  }
  FloatColor operator*(float x) const { return FloatColor(r*x, g*x, b*x); }
  void operator+=(const FloatColor& c) { r+=c.r; g+=c.g; b+=c.b; }
  FloatColor operator+(const FloatColor& c) const { return FloatColor(r+c.r, g+c.g, b+c.b); }
  void Reset() { r=g=b=0.0; }

  float r, g, b;
};


int main(int argc, char **argv) {
  LightBuffer* l = GetLightBuffer();
  std::vector<int> geometry;
  l->GetGeometry(&geometry);

  double z = 0.0;
  int frame = 0;
  struct timeval t = Time();
  for (int i = 0; i < geometry.size(); i++) {
    for (int lamp = 0; lamp < geometry[i]; lamp++) {
      l->Set(i, lamp, Color(0,0,0));
    }
  }
  l->Show();
  l->Show();
  l->Show();
  while (1) {
    int j = 0;
    float m = 255.0/6;
    FloatColor a(m, m, 0);
    FloatColor b(0, m, m);
    FloatColor c(m, 0, m);
    FloatColor d(m, 0, 0);
    FloatColor e(0, m, 0);
    FloatColor f(0, 0, m);
    for (int i = 0; i < geometry.size(); i++) {
      for (int lamp = 0; lamp < geometry[i]; lamp++) {
	++j;
	float Z = z + sin(z + i*5);
	l->Set(i, lamp, (
		 FloatColor(128, 128, 128) +
		 a * sin(Z + j / 8.0) +
		 b * sin(Z * 0.77 + j / 8.0) +
		 c * sin(Z * 0.55 + j / 8.0) +
		 d * sin(Z * 0.33 + j / 8.0) +
		 e * sin(Z * 0.11 + j / 8.0) +
		 f * sin(Z * 0.22 + j / 8.0)).GetColor());
      }
    }
    l->Show();

    struct timeval now = Time();

    z += 0.1 * 30 * (now - t);
    t = now;
    fprintf(stderr, "\rFrame = %d ", frame++);
  }
}
